#include <stdio.h>
#include <vdpau/vdpau_x11.h>
#include "traceplay.h"
#include "common.h"

Display *     x_display;
int           x_screen;
Window        x_window;

VdpDevice           vdp_device;
VdpGetProcAddress * vdp_get_proc_address;

VdpPresentationQueueTarget vdp_flip_target;
VdpPresentationQueue       vdp_flip_queue;

VdpDeviceDestroy *                          vdp_device_destroy;
VdpPresentationQueueTargetDestroy *         vdp_presentation_queue_target_destroy;
VdpPresentationQueueCreate *                vdp_presentation_queue_create;
VdpPresentationQueueDestroy *               vdp_presentation_queue_destroy;
VdpPresentationQueueDisplay *               vdp_presentation_queue_display;
VdpPresentationQueueBlockUntilSurfaceIdle * vdp_presentation_queue_block_until_surface_idle;
VdpPresentationQueueTargetCreateX11 *       vdp_presentation_queue_target_create_x11;

void * traceplay_video_surface_post_create(
    struct Traceplay_Ctx * ctx,
    VdpVideoSurface        surface
)
{
    return 0;
}

void * traceplay_output_surface_post_create(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface
)
{
    return 0;
}

void traceplay_video_surface_pre_destroy(
    struct Traceplay_Ctx * ctx,
    VdpVideoSurface        surface,
    void *                 surface_context
)
{
}

void traceplay_output_surface_pre_destroy(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
}

void traceplay_display_video_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpVideoSurface        surface,
    void *                 surface_context
)
{
    printf("ERROR at %s:%d: display_video_surface called\n", __FILE__, __LINE__);
    common_exit_error();
}

void traceplay_wait_for_output_surface_idle(
    struct Traceplay_Ctx * ctx,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
    VdpTime t;
    vdp_presentation_queue_block_until_surface_idle(vdp_flip_queue, surface, &t);
}

void traceplay_display_output_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpOutputSurface       surface,
    void *                 surface_context
)
{
    vdp_presentation_queue_display(vdp_flip_queue, surface, 0, 0, 0);
}

void traceplay_process_surface(
    struct Traceplay_Ctx * ctx,
    uint32_t               frame_number,
    VdpVideoSurface        vid_surface,
    void *                 vid_surface_context,
    VdpOutputSurface       out_surface,
    void *                 out_surface_context
)
{
    printf("ERROR at %s:%d: process_surface called\n", __FILE__, __LINE__);
    common_exit_error();
}

Traceplay_Callbacks tpcb = {
    traceplay_video_surface_post_create,
    traceplay_output_surface_post_create,
    traceplay_video_surface_pre_destroy,
    traceplay_output_surface_pre_destroy,
    traceplay_display_video_surface,
    traceplay_wait_for_output_surface_idle,
    traceplay_display_output_surface,
    traceplay_process_surface
};

void initX11(void)
{
    unsigned long windowMask;
    XSetWindowAttributes winAttrib;

    // X11 setup

    x_display = XOpenDisplay(NULL);
    if (!x_display) {
        printf("ERROR at %s:%d: XOpenDisplay failed\n", __FILE__, __LINE__);
        common_exit_error();
    }
    x_screen = DefaultScreen(x_display);

    windowMask = CWBorderPixel; 
    winAttrib.border_pixel = BlackPixel(x_display, x_screen);
    winAttrib.override_redirect = 0;

    x_window = XCreateWindow(
        x_display,
        RootWindow(x_display, x_screen),
        100,
        100,
        trace_stream_conf_ptr->out_width,
        trace_stream_conf_ptr->out_height,
        0,
        CopyFromParent,
        InputOutput,
        CopyFromParent,
        windowMask,
        &winAttrib
    );

    XSelectInput(x_display, x_window, StructureNotifyMask);

    XMapRaised(x_display, x_window);

    XFlush(x_display);
}

void initVDPAU(void)
{
    VdpStatus vdp_st;

    // Create Device

    vdp_st = vdp_device_create_x11(
        x_display,
        x_screen,
        &vdp_device,
        &vdp_get_proc_address
    );
    CHECK_ST

    // Entry point retrieval

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_DEVICE_DESTROY,
        (void **)&vdp_device_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_TARGET_DESTROY,
        (void **)&vdp_presentation_queue_target_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_CREATE,
        (void **)&vdp_presentation_queue_create
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_DESTROY,
        (void **)&vdp_presentation_queue_destroy
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_DISPLAY,
        (void **)&vdp_presentation_queue_display
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_BLOCK_UNTIL_SURFACE_IDLE,
        (void **)&vdp_presentation_queue_block_until_surface_idle
    );
    CHECK_ST

    vdp_st = vdp_get_proc_address(
        vdp_device,
        VDP_FUNC_ID_PRESENTATION_QUEUE_TARGET_CREATE_X11,
        (void **)&vdp_presentation_queue_target_create_x11
    );
    CHECK_ST

    vdp_st = vdp_presentation_queue_target_create_x11(
        vdp_device,
        x_window,
        &vdp_flip_target
    );
    CHECK_ST

    vdp_st = vdp_presentation_queue_create(
        vdp_device,
        vdp_flip_target,
        &vdp_flip_queue
    );
    CHECK_ST
}

void finiVDPAU(void)
{
    vdp_presentation_queue_destroy(
        vdp_flip_queue
    );

    vdp_presentation_queue_target_destroy(
        vdp_flip_target
    );

    vdp_device_destroy(
        vdp_device
    );
}

void finiX11(void)
{
    XDestroyWindow(x_display, x_window);
    XCloseDisplay(x_display);
}

void help()
{
    printf("app [options]\n");
    common_print_help();
}

int main(int argc, char * * argv)
{
    int ret;

    app_name = "traceplay_pq";
    support_opt_gold_surface_type_mask =
        (1 << SURFACE_TYPE_VDP_VID_DEC_ORDER) |
        (1 << SURFACE_TYPE_VDP_VID_DISP_ORDER) |
        (1 << SURFACE_TYPE_VDP_OUT);

    common_init();

    argc--;
    argv++;
    while (argc) {
        if ((ret = common_parse_arg(&argc, &argv)) != COMMON_PARSE_ARG_NOT_PARSED) {
            if (ret != COMMON_PARSE_ARG_OK) {
                help();
                common_exit_error();
            }
        }
        else {
            printf("ERROR: Unknown option '%s'\n", argv[0]);
            help();
            common_exit_error();
        }
        argc--;
        argv++;
    }

    if (common_check_args() != COMMON_CHECK_ARGS_OK) {
        help();
        common_exit_error();
    }

    common_load_stream();

    initX11();
    initVDPAU();
    common_playback_init(
        vdp_device,
        vdp_get_proc_address,
        0
    );
    struct Traceplay_Ctx *traceplay_ctx = traceplay_init(
        vdp_device,
        vdp_get_proc_address,
        TRACEPLAY_PLAY_TYPE_OUTPUT,
        &tpcb,
        trace_stream_conf_ptr
    );
    CHECK_PTR(traceplay_ctx)
    traceplay_play(
        traceplay_ctx,
        trace_cmds_ptr,
        opt_repeat_count
    );
    traceplay_fini(
        traceplay_ctx
    );
    common_playback_fini();
    finiVDPAU();
    finiX11();

    common_exit_ok();
    return 0;
}

